# -*- coding: utf-8 -*-
#
#  result.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

class Result(object):
    HIDE = 1
    DONE = 2
    PROMPT = 3
    SUSPEND = 4

    def __init__(self, value):
        self._value = value

    def __int__(self):
        return self._value

    def __lt__(self, other):
        if isinstance(other, Result):
            return int(self) < int(other)
        else:
            return int(self) < other

    def __eq__(self, other):
        if isinstance(other, Result):
            return int(self) == int(other)
        else:
            return int(self) == other

# Easy shortcuts
HIDE = Result(Result.HIDE)
DONE = Result(Result.DONE)

class Prompt(Result):
    def __init__(self, prompt, autocomplete={}):
        Result.__init__(self, Result.PROMPT)

        self.prompt = prompt
        self.autocomplete = autocomplete

class Suspend(Result):
    def __init__(self):
        Result.__init__(self, Result.SUSPEND)
        self._callbacks = []

    def register(self, cb, *args):
        self._callbacks.append([cb, args])

    def resume(self):
        for cb in self._callbacks:
            args = cb[1]
            cb[0](*args)

# ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  format.py - format commander module
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import commander.commands as commands

__commander_module__ = True

def remove_trailing_spaces(view, all=False):
    """Remove trailing spaces: format.remove-trailing-spaces [&lt;all&gt;]

Remove trailing spaces in the selection. If there is no selection, trailing
spaces are removed from the whole document. When the optional argument
&lt;all&gt; is specified, trailing spaces will be removed from all
the open documents."""

    if all:
        buffers = view.get_toplevel().get_documents()
    else:
        buffers = [view.get_buffer()]

    for buf in buffers:
        bounds = buf.get_selection_bounds()

        if not bounds:
            bounds = buf.get_bounds()

        buf.begin_user_action()

        try:
            # For each line, remove trailing spaces
            if not bounds[1].ends_line():
                bounds[1].forward_to_line_end()

            until = buf.create_mark(None, bounds[1], False)
            start = bounds[0]
            start.set_line_offset(0)

            while start.compare(buf.get_iter_at_mark(until)) < 0:
                end = start.copy()
                end.forward_to_line_end()
                last = end.copy()

                if end.equal(buf.get_end_iter()):
                    end.backward_char()

                while end.get_char().isspace() and end.compare(start) > 0:
                    end.backward_char()

                if not end.ends_line():
                    if not end.get_char().isspace():
                        end.forward_char()

                    buf.delete(end, last)

                start = end.copy()
                start.forward_line()

        except Exception as e:
            print(e)

        buf.delete_mark(until)
        buf.end_user_action()

    return commands.result.HIDE

def _transform(view, how, all):
    if all:
        buffers = view.get_toplevel().get_documents()
    else:
        buffers = [view.get_buffer()]

    for buf in buffers:
        bounds = buf.get_selection_bounds()

        if not bounds:
            start = buf.get_iter_at_mark(buf.get_insert())
            end = start.copy()

            if not end.ends_line():
                end.forward_to_line_end()

            bounds = [start, end]

        if not bounds[0].equal(bounds[1]):
            text = how(bounds[0].get_text(bounds[1]))

            buf.begin_user_action()
            buf.delete(bounds[0], bounds[1])
            buf.insert(bounds[0], text)
            buf.end_user_action()

    return commands.result.HIDE

def upper(view, all=False):
    """Make upper case: format.upper [&lt;all&gt;]

Transform text in selection to upper case. If the optional argument &lt;all&gt;
is specified, text in all the open documents will be transformed."""
    return _transform(view, lambda x: x.upper(), all)

def lower(view, all=False):
    """Make lower case: format.lower [&lt;all&gt;]

Transform text in selection to lower case. If the optional argument &lt;all&gt;
is specified, text in all the open documents will be transformed."""
    return _transform(view, lambda x: x.lower(), all)

def title(view, all=False):
    """Make title case: format.title [&lt;all&gt;]

Transform text in selection to title case. If the optional argument &lt;all&gt;
is specified, text in all the open documents will be transformed."""
    return _transform(view, lambda x: x.title().replace('_', ''), all)

# ex:ts=4:et



# -*- coding: utf-8 -*-

#  Copyright (C) 2014 - Garrett Regier
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc.  51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301 USA.

from gi.repository import GLib

import abc
import collections
import queue
import threading
import traceback

from .debug import debug


class WorkerThread(threading.Thread):
    __metaclass__ = abc.ABCMeta

    __sentinel = object()

    def __init__(self, callback, chunk_size=1, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__callback = callback
        self.__chunk_size = chunk_size

        self.__quit = threading.Event()
        self.__has_idle = threading.Event()

        self.__tasks = queue.Queue()
        self.__results = collections.deque()

    @abc.abstractmethod
    def handle_task(self, *args, **kwargs):
        raise NotImplementedError

    # TODO: add, put, push?
    def push(self, *args, **kwargs):
        self.__tasks.put((args, kwargs))

    def __close(self, process_results):
        self.__quit.set()

        # Prevent the queue.get() from blocking forever
        self.__tasks.put(self.__sentinel)

        super().join()

        if not process_results:
            self.__results.clear()

        else:
            while self.__in_idle() is GLib.SOURCE_CONTINUE:
                pass

    def terminate(self):
        self.__close(False)

    def join(self):
        self.__close(True)

    def clear(self):
        old_tasks = self.__tasks
        self.__tasks = queue.Queue(1)

        # Prevent the queue.get() from blocking forever
        old_tasks.put(self.__sentinel)

        # Block until the old queue has finished, otherwise
        # a old result could be added to the new results queue
        self.__tasks.put(self.__sentinel)
        self.__tasks.put(self.__sentinel)

        old_tasks = self.__tasks
        self.__tasks = queue.Queue()

        # Switch to the new queue
        old_tasks.put(self.__sentinel)

        # Finally, we can now create a new deque without
        # the possibility of any old results being added to it
        self.__results.clear()

    def run(self):
        while not self.__quit.is_set():
            task = self.__tasks.get()
            if task is self.__sentinel:
                continue

            args, kwargs = task

            try:
                result = self.handle_task(*args, **kwargs)

            except Exception:
                traceback.print_exc()
                continue

            self.__results.append(result)

            # Avoid having an idle for every result
            if not self.__has_idle.is_set():
                self.__has_idle.set()

                debug('%s<%s>: result callback idle started' %
                      (type(self).__name__, self.name))
                GLib.source_set_name_by_id(GLib.idle_add(self.__in_idle),
                                           '[gedit] git %s result callback idle' %
                                           (type(self).__name__,))

    def __in_idle(self):
        try:
            for i in range(self.__chunk_size):
                result = self.__results.popleft()

                try:
                    self.__callback(result)

                except Exception:
                    traceback.print_exc()

        except IndexError:
            # Must be cleared before we check the results length
            self.__has_idle.clear()

            # Only remove the idle when there are no more items,
            # some could have been added after the IndexError was raised
            if len(self.__results) == 0:
                debug('%s<%s>: result callback idle finished' %
                      (type(self).__name__, self.name))
                return GLib.SOURCE_REMOVE

        return GLib.SOURCE_CONTINUE

# ex:ts=4:et:



# -*- coding: utf-8 -*-

#  Copyright (C) 2013-2014 - Garrett Regier
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc.  51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301 USA.

from gi.repository import GLib, GObject, Gio, Gedit, Ggit

import collections.abc
import weakref

from .appactivatable import GitAppActivatable
from .debug import debug
from .workerthread import WorkerThread


class FileNode(object):
    __slots__ = ('id', 'name')

    def __init__(self, msg):
        self.id = msg.id
        self.name = msg.name


class FileNodes(collections.abc.MutableMapping):
    __slots__ = ('__data')

    def __init__(self):
        self.__data = {}

    def __type_name(self, type):
        if type.__module__ == 'builtins':
            return type.__name__

        return '.'.join((type.__module__, type.__name__))

    def __typecheck(self, value, desired_type):
        if not isinstance(value, desired_type):
            raise TypeError('%s argument expected, got %s.' %
                            (self.__type_name(desired_type),
                             self.__type_name(type(value))))

    def __delitem__(self, location):
        self.__typecheck(location, Gio.File)

        del self.__data[location.get_uri()]

    def __getitem__(self, location):
        self.__typecheck(location, Gio.File)

        return self.__data[location.get_uri()]

    def __setitem__(self, location, node):
        self.__typecheck(location, Gio.File)
        self.__typecheck(node, FileNode)

        if location in self:
            debug('Location is already inserted: %s' % (location.get_uri()))

        self.__data[location.get_uri()] = node

    def __iter__(self):
        return (Gio.File.new_for_uri(x) for x in self.__data)

    def __len__(self):
        return len(self.__data)


class GitStatusThread(WorkerThread):
    def push(self, repo, location):
        if repo is None:
            debug('Invalid repository', print_stack=True)
            return
 
        git_dir = repo.get_location().get_uri()
        if location.get_uri().startswith(git_dir):
            debug('Invalid location: "%s" is in git dir "%s"' %
                  (location.get_uri(), git_dir), print_stack=True)
            return

        workdir = repo.get_workdir()
        if workdir.get_relative_path(location) is None:
            debug('Invalid location "%s" for workdir "%s"' %
                  (location.get_uri(), workdir.get_uri()), print_stack=True)
            return

        super().push(repo, location)

    def handle_task(self, repo, location):
        return location, repo.file_status(location)


class GitWindowActivatable(GObject.Object, Gedit.WindowActivatable):
    window = GObject.Property(type=Gedit.Window)

    windows = weakref.WeakValueDictionary()

    def __init__(self):
        super().__init__()

        self.view_activatables = weakref.WeakSet()

    @classmethod
    def register_view_activatable(cls, view_activatable):
        window = view_activatable.view.get_toplevel()

        if window not in cls.windows:
            return None

        window_activatable = cls.windows[window]

        window_activatable.view_activatables.add(view_activatable)
        view_activatable.connect('notify::status',
                                 window_activatable.notify_status)

        return window_activatable

    def do_activate(self):
        # self.window is not set until now
        self.windows[self.window] = self

        self.app_activatable = GitAppActivatable.get_instance()

        self.bus = self.window.get_message_bus()

        self.git_status_thread = GitStatusThread(self.update_location)
        self.git_status_thread.start()

        self.file_nodes = FileNodes()
        self.monitors = {}
        self.has_focus = True

        self.gobject_signals = {
            self.window: [
                self.window.connect('tab-removed', self.tab_removed),
                self.window.connect('focus-in-event', self.focus_in_event),
                self.window.connect('focus-out-event', self.focus_out_event)
            ],

            # GeditMessageBus.connect() shadows GObject.connect()
            self.bus: [
                GObject.Object.connect(self.bus, 'unregistered',
                                       self.unregistered)
            ]
        }

        # It is safe to connect to these even
        # if the file browser is not enabled yet
        self.bus_signals = [
            self.bus.connect('/plugins/filebrowser', 'root_changed',
                             self.root_changed, None),
            self.bus.connect('/plugins/filebrowser', 'inserted',
                             self.inserted, None),
            self.bus.connect('/plugins/filebrowser', 'deleted',
                             self.deleted, None)
        ]

        self.refresh()

    def do_deactivate(self):
        self.clear_monitors()
        self.git_status_thread.terminate()

        for gobject, sids in self.gobject_signals.items():
            for sid in sids:
                # GeditMessageBus.disconnect() shadows GObject.disconnect()
                GObject.Object.disconnect(gobject, sid)

        for sid in self.bus_signals:
            self.bus.disconnect(sid)

        self.file_nodes = FileNodes()
        self.gobject_signals = {}
        self.bus_signals = []

        self.refresh()

    def refresh(self):
        if self.bus.is_registered('/plugins/filebrowser', 'refresh'):
            self.bus.send('/plugins/filebrowser', 'refresh')

    def get_view_activatable_by_view(self, view):
        for view_activatable in self.view_activatables:
            if view_activatable.view == view:
                return view_activatable

        return None

    def get_view_activatable_by_location(self, location):
        for view_activatable in self.view_activatables:
            buf = view_activatable.view.get_buffer()
            if buf is None:
                continue

            view_location = buf.get_file().get_location()
            if view_location is None:
                continue

            if view_location.equal(location):
                return view_activatable

        return None

    def notify_status(self, view_activatable, psepc):
        location = view_activatable.view.get_buffer().get_file().get_location()
        if location is None:
            return

        if location not in self.file_nodes:
            return

        repo = self.get_repository(location)
        if repo is not None:
            self.git_status_thread.push(repo, location)

    def tab_removed(self, window, tab):
        view = tab.get_view()

        # Need to remove the view activatable otherwise update_location()
        # might use the view's status and not the file's actual status
        view_activatable = self.get_view_activatable_by_view(view)
        if view_activatable is not None:
            self.view_activatables.remove(view_activatable)

        location = view.get_buffer().get_file().get_location()
        if location is None:
            return

        if location not in self.file_nodes:
            return

        repo = self.get_repository(location)
        if repo is not None:
            self.git_status_thread.push(repo, location)

    def focus_in_event(self, window, event):
        # Enables the file monitors so they can cause things
        # to update again. We disabled them when the focus
        # was lost and we will instead do a full update now.
        self.has_focus = True

        self.app_activatable.clear_repositories()

        for view_activatable in self.view_activatables:
            # Must reload the location's contents, not just rediff
            GLib.idle_add(view_activatable.update_location)

        for location in self.file_nodes:
            # Still need to update the git status
            # as the file could now be in .gitignore
            repo = self.get_repository(location)
            if repo is not None:
                self.git_status_thread.push(repo, location)

    def focus_out_event(self, window, event):
        # Disables the file monitors so they don't
        # cause anything to update. We will do a
        # full update when we have focus again.
        self.has_focus = False

    def unregistered(self, bus, object_path, method):
        # Avoid warnings like crazy if the file browser becomes disabled
        if object_path == '/plugins/filebrowser' and method == 'root_changed':
            self.clear_monitors()
            self.git_status_thread.clear()
            self.file_nodes = FileNodes()

    def get_repository(self, location, is_dir=False):
        return self.app_activatable.get_repository(location, is_dir)

    def root_changed(self, bus, msg, data=None):
        self.clear_monitors()
        self.git_status_thread.clear()
        self.file_nodes = FileNodes()

        location = msg.location

        repo = self.get_repository(location, True)
        if repo is not None:
            self.monitor_directory(location)

    def inserted(self, bus, msg, data=None):
        location = msg.location

        repo = self.get_repository(location, msg.is_directory)
        if repo is None:
            return

        if msg.is_directory:
            self.monitor_directory(location)

        else:
            self.file_nodes[location] = FileNode(msg)
            self.git_status_thread.push(repo, location)

    def deleted(self, bus, msg, data=None):
        location = msg.location
        uri = location.get_uri()

        if uri in self.monitors:
            self.monitors[uri].cancel()
            del self.monitors[uri]

        else:
            try:
                del self.file_nodes[location]

            except KeyError:
                pass

    def update_location(self, result):
        location, status = result

        # The node may have been deleted
        # before the status was determined
        try:
            file_node = self.file_nodes[location]

        except KeyError:
            return

        if status is None or not status & Ggit.StatusFlags.IGNORED:
            view_activatable = self.get_view_activatable_by_location(location)
            if view_activatable is not None:
                status = view_activatable.status

        markup = GLib.markup_escape_text(file_node.name)

        if status is not None:
            if status & Ggit.StatusFlags.INDEX_NEW or \
                    status & Ggit.StatusFlags.WORKING_TREE_NEW or \
                    status & Ggit.StatusFlags.INDEX_MODIFIED or \
                    status & Ggit.StatusFlags.WORKING_TREE_MODIFIED:
                markup = '<span weight="bold">%s</span>' % (markup)

            elif status & Ggit.StatusFlags.INDEX_DELETED or \
                    status & Ggit.StatusFlags.WORKING_TREE_DELETED:
                markup = '<span strikethrough="true">%s</span>' % (markup)

        self.bus.send_sync('/plugins/filebrowser', 'set_markup',
                           id=file_node.id, markup=markup)

    def clear_monitors(self):
        for uri in self.monitors:
            self.monitors[uri].cancel()

        self.monitors = {}

    def monitor_directory(self, location):
        try:
            monitor = location.monitor(Gio.FileMonitorFlags.NONE, None)

        except GLib.Error as e:
            debug('Failed to monitor directory "%s": %s' %
                  (location.get_uri(), e))
            return

        self.monitors[location.get_uri()] = monitor
        monitor.connect('changed', self.monitor_changed)

    def monitor_changed(self, monitor, file_a, file_b, event_type):
        # Don't update anything as we will do
        # a full update when we have focus again
        if not self.has_focus:
            return

        # Only monitor for changes as the file browser
        # will emit signals for the other event types
        if event_type != Gio.FileMonitorEvent.CHANGED:
            return

        for f in (file_a, file_b):
            if f is None:
                continue

            # Must let the view activatable know
            # that its location's contents have changed
            view_activatable = self.get_view_activatable_by_location(f)
            if view_activatable is not None:
                # Must reload the location's contents, not just rediff
                GLib.idle_add(view_activatable.update_location)

                # Still need to update the git status
                # as the file could now be in .gitignore

            if f in self.file_nodes:
                repo = self.get_repository(f)
                if repo is not None:
                    self.git_status_thread.push(repo, f)

# ex:ts=4:et:



# -*- coding: utf-8 -*-
#
#  move.py - move commander module
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import commander.commands as commands
from gi.repository import Gdk, Gtk
import re

__commander_module__ = True

def _move(view, what, num, modifier):
    try:
        num = int(num)
    except:
        raise commands.exceptions.Execute('Invalid number: ' + str(num))

    view.emit('move-cursor', what, num, modifier & Gdk.EventMask.CONTROL_MASK)
    return commands.result.HIDE

def word(view, modifier, num=1):
    """Move cursor per word: move.word &lt;num&gt;

Move the cursor per word (use negative num to move backwards)"""
    return _move(view, Gtk.MovementStep.WORDS, num, modifier)

def line(view, modifier, num=1):
    """Move cursor per line: move.line &lt;num&gt;

Move the cursor per line (use negative num to move backwards)"""
    return _move(view, Gtk.MovementStep.DISPLAY_LINES, num, modifier)

def char(view, modifier, num=1):
    """Move cursor per char: move.char &lt;num&gt;

Move the cursor per char (use negative num to move backwards)"""
    return _move(view, Gtk.MovementStep.VISUAL_POSITIONS, num, modifier)

def paragraph(view, modifier, num=1):
    """Move cursor per paragraph: move.paragraph &lt;num&gt;

Move the cursor per paragraph (use negative num to move backwards)"""
    return _move(view, Gtk.MovementStep.PARAGRAPHS, num, modifier)

def regex(view, modifier, regex, num=1):
    """Move cursor per regex: move.regex &lt;num&gt;

Move the cursor per regex (use negative num to move backwards)"""
    try:
        r = re.compile(regex, re.DOTALL | re.MULTILINE | re.UNICODE)
    except Exception as e:
        raise commands.exceptions.Execute('Invalid regular expression: ' + str(e))

    try:
        num = int(num)
    except Exception as e:
        raise commands.exceptions.Execute('Invalid number: ' + str(e))

    buf = view.get_buffer()
    start = buf.get_iter_at_mark(buf.get_insert())

    if num > 0:
        end = buf.get_end_iter()
    else:
        end = start.copy()
        start = buf.get_start_iter()

    text = start.get_text(end)
    ret = list(r.finditer(text))

    if num < 0:
        ret.reverse()

    idx = min(abs(num), len(ret))

    if idx > 0:
        found = ret[idx - 1]
        start = buf.get_iter_at_mark(buf.get_insert())

        if num < 0:
            start.backward_char(len(text) - found.start(0))
        else:
            start.forward_chars(found.start(0))

        if modifier & Gdk.EventMask.CONTROL_MASK:
            buf.move_mark(buf.get_selection_bound(), start)
        else:
            buf.move_mark(buf.get_insert(), start)
            buf.move_mark(buf.get_selection_bound(), start)

        visible = view.get_visible_rect()
        loc = view.get_iter_location(start)

        # Scroll there if needed
        if loc.y + loc.height < visible.y or loc.y > visible.y + visible.height:
            view.scroll_to_mark(buf.get_insert(), 0.2, True, 0, 0.5)

    return commands.result.HIDE

# ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  finder.py - finder commander module
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from xml.sax import saxutils
import commander.commands.result
import commander.utils as utils
from gi.repository import Gdk, Gtk

class Finder:
    FIND_STARTMARK = 'gedit-commander-find-startmark'
    FIND_ENDMARK = 'gedit-commander-find-endmark'

    FIND_RESULT_STARTMARK = 'gedit-commander-find-result-startmark'
    FIND_RESULT_ENDMARK = 'gedit-commander-find-result-endmark'

    def __init__(self, entry):
        self.entry = entry
        self.view = entry.view()

        self.findstr = None
        self.replacestr = None

        self.search_boundaries = utils.Struct({'start': None, 'end': None})
        self.find_result = utils.Struct({'start': None, 'end': None})

        self.unescapes = [
            ['\\n', '\n'],
            ['\\r', '\r'],
            ['\\t', '\t']
        ]

        self.from_start = False
        self.search_start_mark = None

    def unescape(self, s):
        for esc in self.unescapes:
            s = s.replace(esc[0], esc[1])

        return s

    def do_find(self, bounds):
        return None

    def get_replace(self, text):
        return self.replacestr

    def set_replace(self, replacestr):
        self.replacestr = self.unescape(replacestr)

    def set_find(self, findstr):
        self.findstr = self.unescape(findstr)

    def get_find(self):
        return self.findstr

    def select_last_result(self):
        buf = self.view.get_buffer()

        startiter = buf.get_iter_at_mark(self.find_result.start)
        enditer = buf.get_iter_at_mark(self.find_result.end)

        buf.select_range(startiter, enditer)

        visible = self.view.get_visible_rect()
        loc = self.view.get_iter_location(startiter)

        # Scroll there if needed
        if loc.y + loc.height < visible.y or loc.y > visible.y + visible.height:
            self.view.scroll_to_iter(startiter, 0.2, True, 0, 0.5)

    def find_next(self, select=False):
        buf = self.view.get_buffer()

        # Search from the end of the last result to the end of the search boundary
        bounds = [buf.get_iter_at_mark(self.find_result.end),
                  buf.get_iter_at_mark(self.search_boundaries.end)]

        ret = self.do_find(bounds)

        # Check if we need to wrap around if nothing is found
        if self.search_start_mark:
            startiter = buf.get_iter_at_mark(self.search_start_mark)
        else:
            startiter = None

        startbound = buf.get_iter_at_mark(self.search_boundaries.start)

        if not ret and not self.from_start and (startiter and not startiter.equal(startbound)):
            self.from_start = True

            # Try from beginning
            bounds[0] = buf.get_start_iter()
            bounds[1] = startiter

            # Make sure to just stop at the start of the previous
            self.search_boundaries.end = self.search_start_mark

            ret = self.do_find(bounds)

        if not ret:
            return False
        else:
            # Mark find result
            buf.move_mark(self.find_result.start, ret[0])
            buf.move_mark(self.find_result.end, ret[1])

            if select:
                self.select_last_result()

            return True

    def _create_or_move(self, markname, piter, left_gravity):
        buf = self.view.get_buffer()
        mark = buf.get_mark(markname)

        if not mark:
            mark = buf.create_mark(markname, piter, left_gravity)
        else:
            buf.move_mark(mark, piter)

        return mark

    def find_first(self, doend=True, select=False):
        words = []
        buf = self.view.get_buffer()

        while not self.findstr:
            fstr, words, modifier = (yield commander.commands.result.Prompt('Find:'))

            if fstr:
                self.set_find(fstr)

        # Determine search area
        bounds = list(buf.get_selection_bounds())

        if self.search_start_mark:
            buf.delete_mark(self.search_start_mark)
            self.search_start_mark = None

        if not bounds:
            # Search in the whole buffer, from the current cursor position on to the
            # end, and then continue to start from the beginning of the buffer if needed
            bounds = list(buf.get_bounds())
            self.search_start_mark = buf.create_mark(None, buf.get_iter_at_mark(buf.get_insert()), True)
            selection = False
        else:
            selection = True

        bounds[0].order(bounds[1])

        # Set marks at the boundaries
        self.search_boundaries.start = self._create_or_move(Finder.FIND_STARTMARK, bounds[0], True)
        self.search_boundaries.end = self._create_or_move(Finder.FIND_ENDMARK, bounds[1], False)

        # Set the result marks so the next find will start at the correct location
        if selection:
            piter = bounds[0]
        else:
            piter = buf.get_iter_at_mark(buf.get_insert())

        self.find_result.start = self._create_or_move(Finder.FIND_RESULT_STARTMARK, piter, True)
        self.find_result.end = self._create_or_move(Finder.FIND_RESULT_ENDMARK, piter, False)

        if not self.find_next(select=select):
            if doend:
                self.entry.info_show('<i>Search hit end of the document</i>', True)

            yield commander.commands.result.DONE
        else:
            yield True

    def cancel(self):
        buf = self.view.get_buffer()

        buf.set_search_text('', 0)
        buf.move_mark(buf.get_selection_bound(), buf.get_iter_at_mark(buf.get_insert()))

        if self.search_start_mark:
            buf.delete_mark(self.search_start_mark)

    def find(self, findstr):
        if findstr:
            self.set_find(findstr)

        buf = self.view.get_buffer()

        try:
            if (yield self.find_first(select=True)):
                while True:
                    argstr, words, modifier = (yield commander.commands.result.Prompt('Search next [<i>%s</i>]:' % (saxutils.escape(self.findstr),)))

                    if argstr:
                        self.set_find(argstr)

                    if not self.find_next(select=True):
                        break

                self.entry.info_show('<i>Search hit end of the document</i>', True)
        except GeneratorExit as e:
            self.cancel()
            raise e

        self.cancel()
        yield commander.commands.result.DONE

    def _restore_cursor(self, mark):
        buf = mark.get_buffer()

        buf.place_cursor(buf.get_iter_at_mark(mark))
        buf.delete_mark(mark)

        self.view.scroll_to_mark(buf.get_insert(), 0.2, True, 0, 0.5)

    def get_current_replace(self):
        buf = self.view.get_buffer()
        bounds = utils.Struct({'start': buf.get_iter_at_mark(self.find_result.start),
                               'end': buf.get_iter_at_mark(self.find_result.end)})

        if not bounds.start.equal(bounds.end):
            text = bounds.start.get_text(bounds.end)
            return self.get_replace(text)
        else:
            return self.replacestr

    def replace(self, findstr, replaceall=False, replacestr=None):
        if findstr:
            self.set_find(findstr)

        if replacestr != None:
            self.set_replace(replacestr)

        # First find something
        buf = self.view.get_buffer()

        if replaceall:
            startmark = buf.create_mark(None, buf.get_iter_at_mark(buf.get_insert()), False)

        ret = (yield self.find_first(select=not replaceall))

        if not ret:
            yield commander.commands.result.DONE

        self.scroll_back = False

        # Then ask for the replacement string
        if not self.replacestr:
            try:
                if replaceall:
                    self.scroll_back = True
                    self.select_last_result()

                replacestr, words, modifier = (yield commander.commands.result.Prompt('Replace with:'))
                self.set_replace(replacestr)
            except GeneratorExit as e:
                if replaceall:
                    self._restore_cursor(startmark)

                self.cancel()
                raise e

        # On replace all, wrap it in begin/end user action
        if replaceall:
            buf.begin_user_action()

        try:
            while True:
                if not replaceall:
                    rep, words, modifier = (yield commander.commands.result.Prompt('Replace next [%s]:' % (saxutils.escape(self.get_current_replace()),)))

                    if rep:
                        self.set_replace(rep)

                bounds = utils.Struct({'start': buf.get_iter_at_mark(self.find_result.start),
                                       'end': buf.get_iter_at_mark(self.find_result.end)})

                # If there is a selection, replace it with the replacement string
                if not bounds.start.equal(bounds.end) and (replaceall or not (modifier & Gdk.ModifierType.CONTROL_MASK)):
                    text = bounds.start.get_text(bounds.end)
                    repl = self.get_replace(text)

                    buf.begin_user_action()
                    buf.delete(bounds.start, bounds.end)
                    buf.insert(bounds.start, repl)
                    buf.end_user_action()

                # Find next
                if not self.find_next(select=not replaceall):
                    if not replaceall:
                        self.entry.info_show('<i>Search hit end of the document</i>', True)

                    break

        except GeneratorExit as e:
            if replaceall:
                self._restore_cursor(startmark)
                buf.end_user_action()

            self.cancel()
            raise e

        if replaceall:
            if self.scroll_back:
                self._restore_cursor(startmark)

            buf.end_user_action()

        self.cancel()
        yield commander.commands.result.DONE

# ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  Copyrignt (C) 2017 Jordi Mas <jmas@softcatala.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import urllib.request, urllib.parse, urllib.error
import json
from .service import Service

try:
    import gettext
    gettext.bindtextdomain('gedit-plugins')
    gettext.textdomain('gedit-plugins')
    _ = gettext.gettext
except:
    _ = lambda s: s


class Yandex(Service):

    g_language_codes = []
    g_language_names = []
    g_locales_names = {}

    DEFAULT_LANGUAGE_NAMES = ["Spanish -> English",
            "English -> Spanish",
            "Catalan -> English",
            "English -> Catalan"]

    DEFAULT_LANGUAGE_CODES = ["es|en",
            "en|es",
            "ca|en",
            "en|ca",
    ]

    SERVER = "https://translate.yandex.net/api/v1.5/tr.json"

    g_language_codes = []
    g_language_names = []

    @staticmethod
    def _clean_for_ut():
        Yandex.g_language_codes = []
        Yandex.g_language_names = []
        Yandex.g_locales_names = {}

    def get_default_language_codes(self):
        return 'en|es'

    def get_api_hint(self):
        return _("You need to obtain an API key at <a href='https://tech.yandex.com/translate/'>https://tech.yandex.com/translate/</a>")

    def has_api_key(self):
        return True

    def set_api_key(self, key):
        self._key = key

    def set_server(self, server):
        pass

    def init(self):
        self._fetch_remote_language_names()

    def get_language_names(self):
        if len(Yandex.g_language_codes) > 0 and len(Yandex.g_language_names) > 0:
            return Yandex.g_language_names

        return self.DEFAULT_LANGUAGE_NAMES

    def get_language_codes(self):
        if len(Yandex.g_language_codes) > 0 and len(Yandex.g_language_names) > 0:
            return Yandex.g_language_codes

        return self.DEFAULT_LANGUAGE_CODES

    def get_language_pair_name(self, source, target, locales_names=None):
        if locales_names is None:
            locales_names = Yandex.g_locales_names

        source = self._get_language_name(source, locales_names)
        target = self._get_language_name(target, locales_names)
        return "{0} -> {1}".format(source, target)

    def _get_language_name(self, langcode, locales_names):
        return locales_names[langcode]

    def _fetch_remote_language_names(self):

        try:

            if len(self._key) == 0:
                return

            if (len(Yandex.g_locales_names) > 0 and
               len(Yandex.g_language_names) > 0 and
               len(Yandex.g_language_codes) > 0):
                    return

            url = "{0}/getLangs?ui=en&key={1}".format(self.SERVER, self._key)
            response = urllib.request.urlopen(url)
            payload = json.loads(response.read().decode("utf-8"))

            language_codes = payload['dirs']
            language_codes = [x.replace('-', '|') for x in language_codes]
            locales_names = payload['langs']

            language_names = []
            for lang_pair in language_codes:
                langs = lang_pair.split('|')
                source = langs[0]
                target = langs[1]
                name = self.get_language_pair_name(source, target, locales_names)
                language_names.append(name)

            Yandex.g_locales_names = locales_names
            Yandex.g_language_names = language_names
            Yandex.g_language_codes = language_codes

        except Exception as e:
            print("_fetch_remote_language_names exception {0}".format(e))

    def translate_text(self, text, language_pair):
        language_pair = language_pair.replace('|', '-')
        url = "{0}/translate?lang={1}&format=plain&key={2}".format(self.SERVER, language_pair, self._key)
        url += "&text=" + urllib.parse.quote_plus(text.encode('utf-8'))
        response = urllib.request.urlopen(url)
        r = response.read().decode("utf-8")
        data = json.loads(r)
        all_text = ''

        texts = data['text']
        for text in texts:
            all_text += text
        return all_text



# -*- coding: utf-8 -*-
#
#  info.py - commander
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

from gi.repository import Pango, Gdk, Gtk
import math

class ScrolledWindow(Gtk.ScrolledWindow):
    __gtype_name__ = "CommanderScrolledWindow"

    def __init__(self):
        Gtk.ScrolledWindow.__init__(self)

        self._max_height = 0
        self._max_lines = 10

        self.view = Gtk.TextView()
        self.view.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self.view.set_editable(False)
        self.view.set_can_focus(False)

        self.view.connect('style-updated', self._on_style_updated)

        self._update_max_height()

        self.view.show()

        self.add(self.view)

        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)

    def _update_max_height(self):
        layout = self.view.create_pango_layout('Some text to measure')
        extents = layout.get_pixel_extents()

        maxh = extents[1].height * self._max_lines

        if maxh != self._max_height:
            self._max_height = maxh
            self.queue_resize()

    def _on_style_updated(self, widget):
        self._update_max_height()

    def do_get_preferred_height(self):
        hp, vp = self.get_policy()

        ret = self.view.get_preferred_height()

        if vp == Gtk.PolicyType.NEVER and ret[0] > self._max_height:
            self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS)
            self.set_min_content_height(self._max_height)
        elif vp == Gtk.PolicyType.ALWAYS and ret[0] < self._max_height:
            self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
            self.set_min_content_height(0)

        return Gtk.ScrolledWindow.do_get_preferred_height(self)

class Info(Gtk.Box):
    __gtype_name__ = "CommanderInfo"

    def __init__(self):
        super(Info, self).__init__()

        self._button_bar = None
        self._status_label = None

        self._build_ui()

    def _build_ui(self):
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.set_spacing(3)
        self.set_can_focus(False)

        self._sw = ScrolledWindow()
        self._sw.set_border_width(6)

        css = Gtk.CssProvider()
        css.load_from_data(bytes("""
.trough {
    background: transparent;
}
""", 'utf-8'))

        self._sw.get_vscrollbar().get_style_context().add_provider(css, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        self._sw.show()
        self.add(self._sw)

        self._attr_map = {
            Pango.AttrType.STYLE: ('style', Pango.AttrInt),
            Pango.AttrType.WEIGHT: ('weight', Pango.AttrInt),
            Pango.AttrType.VARIANT: ('variant', Pango.AttrInt),
            Pango.AttrType.STRETCH: ('stretch', Pango.AttrInt),
            Pango.AttrType.SIZE: ('size', Pango.AttrInt),
            Pango.AttrType.FOREGROUND: ('foreground', Pango.AttrColor),
            Pango.AttrType.BACKGROUND: ('background', Pango.AttrColor),
            Pango.AttrType.UNDERLINE: ('underline', Pango.AttrInt),
            Pango.AttrType.STRIKETHROUGH: ('strikethrough', Pango.AttrInt),
            Pango.AttrType.RISE: ('rise', Pango.AttrInt),
            Pango.AttrType.SCALE: ('scale', Pango.AttrFloat)
        }

    @property
    def text_view(self):
        return self._sw.view

    @property
    def is_empty(self):
        buf = self.text_view.get_buffer()
        return buf.get_start_iter().equal(buf.get_end_iter())

    def status(self, text=None):
        if self._status_label == None and text != None:
            self._status_label = Gtk.Label()

            context = self.text_view.get_style_context()
            state = context.get_state()
            font_desc = context.get_font(state)

            self._status_label.override_font(font_desc)

            self._status_label.override_color(Gtk.StateFlags.NORMAL, context.get_color(Gtk.StateFlags.NORMAL))
            self._status_label.show()
            self._status_label.set_alignment(0, 0.5)
            self._status_label.set_padding(10, 0)
            self._status_label.set_use_markup(True)
            self._status_label.set_halign(Gtk.Align.FILL)

            self._ensure_button_bar()
            self._button_bar.pack_start(self._status_label, True, True, 0)

        if text != None:
            self._status_label.set_markup(text)
        elif self._status_label:
            self._status_label.destroy()

            if not self._button_bar and self.is_empty:
                self.destroy()

    def _attr_to_tag(self, attr):
        buf = self.text_view.get_buffer()
        table = buf.get_tag_table()
        ret = []

        tp = attr.klass.type

        if not tp in self._attr_map:
            return None

        propname, klass = self._attr_map[tp]

        # The hack! Set __class__ so we can access .value/.color,
        # then set __class__ back. This is expected to break at any time
        cls = attr.__class__
        attr.__class__ = klass

        if tp == Pango.AttrType.FOREGROUND or tp == Pango.AttrType.BACKGROUND:
            value = attr.color
        else:
            value = attr.value

        attr.__class__ = cls

        tagname = str(tp) + ':' + str(value)

        tag = table.lookup(tagname)

        if not tag:
            tag = buf.create_tag(tagname)
            tag.set_property(propname, value)

        return tag

    def add_lines(self, line, use_markup=False):
        buf = self.text_view.get_buffer()

        if not buf.get_start_iter().equal(buf.get_end_iter()):
            line = "\n" + line

        if not use_markup:
            buf.insert(buf.get_end_iter(), line)
            return

        try:
            ret = Pango.parse_markup(line, -1, '\x00')
        except Exception as e:
            print('Could not parse markup:', e)
            buf.insert(buf.get_end_iter(), line)
            return

        text = ret[2]

        mark = buf.create_mark(None, buf.get_end_iter(), True)
        buf.insert(buf.get_end_iter(), text)

        attrs = []
        ret[1].filter(lambda x: attrs.append(x))

        for attr in attrs:
            # Try/catch everything since the _attr_to_tag stuff is a big
            # hack
            try:
                tag = self._attr_to_tag(attr)
            except:
                continue

            if not tag is None:
                start = buf.get_iter_at_mark(mark)
                end = start.copy()

                start.forward_chars(attr.start_index)
                end.forward_chars(attr.end_index)

                buf.apply_tag(tag, start, end)

    def add_action(self, name, callback, data=None):
        image = Gtk.Image.new_from_icon_name(name, Gtk.IconSize.MENU)
        image.show()

        self._ensure_button_bar()

        ev = Gtk.EventBox()
        ev.set_visible_window(True)
        ev.add(image)
        ev.show()

        ev.set_halign(Gtk.Align.END)

        self._button_bar.pack_end(ev, False, False, 0)
        ev.get_window().set_cursor(Gdk.Cursor.new(Gdk.CursorType.HAND2))

        ev.connect('button-press-event', self._on_action_activate, callback, data)
        ev.connect('enter-notify-event', self._on_action_enter_notify)
        ev.connect('leave-notify-event', self._on_action_leave_notify)

        ev.connect_after('destroy', self._on_action_destroy)
        return ev

    def clear(self):
        self.text_view.get_buffer().set_text('')

    def _ensure_button_bar(self):
        if not self._button_bar:
            self._button_bar = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=3)
            self._button_bar.show()
            self.pack_start(self._button_bar, False, False, 0)

    def _on_action_destroy(self, widget):
        if self._button_bar and len(self._button_bar.get_children()) == 0:
            self._button_bar.destroy()
            self._button_bar = None

    def _on_action_enter_notify(self, widget, evnt):
        img = widget.get_child()
        img.set_state(Gtk.StateType.PRELIGHT)

    def _on_action_leave_notify(self, widget, evnt):
        img = widget.get_child()
        img.set_state(Gtk.StateType.NORMAL)

    def _on_action_activate(self, widget, evnt, callback, data):
        if data:
            callback(data)
        else:
            callback()

# ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  bookmark.py - bookmark interface commander module
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import commander.commands
import commander.commands.exceptions

__commander_module__ = True

def check_bookmark_plugin(window):
    if not window.get_message_bus().is_registered('/plugins/bookmarks', 'toggle'):
        raise commander.commands.exceptions.Execute("The bookmarks plugin is not installed, not active or too old")

def __default__(view, window):
    """Commander interface to the bookmarks plugin: bookmark

This module provides an interface to the bookmarks plugin from gedit-plugins.
If installed and active, you can add/remove/toggle bookmarks using the
commander."""

    check_bookmark_plugin(window)

    window.get_message_bus().send('/plugins/bookmarks', 'toggle', view=view)

def add(view, window):
    """Add bookmark: bookmark.add

Add bookmark on the current line. If there already is a bookmark on the current
line, nothing happens."""

    check_bookmark_plugin(window)

    window.get_message_bus().send('/plugins/bookmarks', 'add', view=view)

def remove(view, window):
    """Remove bookmark: bookmark.remove

Remove bookmark from the current line. If there is no bookmark on the current
line, nothing happens."""

    check_bookmark_plugin(window)

    window.get_message_bus().send('/plugins/bookmarks', 'remove', view=view)

def toggle(view, window):
    """Toggle bookmark: bookmark.toggle

Toggle bookmark on the current line."""

    check_bookmark_plugin(window)

    window.get_message_bus().send('/plugins/bookmarks', 'toggle', view=view)

def next(view, window):
    """Goto next bookmark: bookmark.next

Jump to the next bookmark location"""

    check_bookmark_plugin(window)

    window.get_message_bus().send('/plugins/bookmarks', 'goto_next', view=view)

def previous(view, window):
    """Goto previous bookmark: bookmark.previous

Jump to the previous bookmark location"""

    check_bookmark_plugin(window)

    window.get_message_bus().send('/plugins/bookmarks', 'goto_previous', view=view)

# ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  regex.py - regex commander module
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import commander.commands as commands
from . import finder

import re

__commander_module__ = True
__root__ = ['regex_i']

class RegexFinder(finder.Finder):
    def __init__(self, entry, flags = 0):
        finder.Finder.__init__(self, entry)

        self.flags = re.UNICODE | re.MULTILINE | flags
        self.groupre = re.compile('(\\\\)?\\$([0-9]+|{(([0-9]+):([^}]+))})')

    def set_find(self, findstr):
        finder.Finder.set_find(self, findstr)

        try:
            self.findre = re.compile(findstr, self.flags)
        except Exception as e:
            raise commands.exceptions.Execute('Invalid regular expression: ' + str(e))

    def do_find(self, bounds):
        buf = self.view.get_buffer()

        text = bounds[0].get_text(bounds[1])
        ret = self.findre.search(text)

        if ret:
            start = bounds[0].copy()
            start.forward_chars(ret.start())

            end = bounds[0].copy()
            end.forward_chars(ret.end())

            return [start, end]
        else:
            return False

    def _transform(self, text, trans):
        if not trans:
            return text

        transforms = {
            'u': lambda x: "%s%s" % (x[0].upper(), x[1:]),
            'U': lambda x: x.upper(),
            'l': lambda x: "%s%s" % (x[0].lower(), x[1:]),
            'L': lambda x: x.lower(),
            't': lambda x: x.title()
        }

        for i in trans.split(','):
            if i in transforms:
                text = transforms[i](text)

        return text

    def _do_re_replace_group(self, matchit, group):
        if group.group(3):
            num = int(group.group(4))
        else:
            num = int(group.group(2))

        if group.group(1):
            return group.group(2)
        elif num < len(matchit.groups()) + 1:
            return self._transform(matchit.group(num), group.group(5))
        else:
            return group.group(0)

    def _do_re_replace(self, matchit):
        return self.groupre.sub(lambda x: self._do_re_replace_group(matchit, x), self.replacestr)

    def get_replace(self, text):
        try:
            return self.findre.sub(self._do_re_replace, text)
        except Exception as e:
            raise commands.exceptions.Execute('Invalid replacement: ' + str(e))

class SemanticFinder(RegexFinder):
    def __init__(self, entry):
        RegexFinder.__init__(self, entry, re.IGNORECASE)

    def split_semantic(self, s):
        # Can be specified in _ separated syntax, or in TitleCase
        if '_' in s:
            return s.lower().split('_')
        else:
            return [m.group(0).lower() for m in re.finditer('[A-Z]+([^A-Z]+)?', s)]

    def set_find(self, findstr):
        self.findparts = self.split_semantic(findstr)
        RegexFinder.set_find(self, '(' + ')(_?)('.join(map(lambda x: re.escape(x), self.findparts)) + ')')

    def set_replace(self, replstr):
        self.replaceparts = self.split_semantic(replstr)
        RegexFinder.set_replace(self, replstr)

    def copy_case(self, orig, repl):
        lo = len(orig)
        lr = len(repl)

        ret = ''

        for i in range(min(lr, lo)):
            if orig[i].isupper():
                ret += repl[i].upper()
            else:
                ret += repl[i].lower()

        if lr > lo:
            if orig.isupper():
                ret += repl[lo:].upper()
            else:
                ret += repl[lo:].lower()

        return ret

    def get_replace(self, text):
        m = self.findre.match(text)
        groups = m.groups()

        ret = []

        for i in range(len(groups)):
            ri = i // 2

            if i % 2 == 0:
                if ri >= len(self.replaceparts):
                    break

                ret.append(self.copy_case(groups[i], self.replaceparts[ri]))
            else:
                ret.append(groups[i])

        return ''.join(ret)

def __default__(entry, argstr):
    """Find regex in document: find.regex &lt;regex&gt;

Find text in the document that matches a given regular expression. The regular
expression syntax is that of python regular expressions."""
    fd = RegexFinder(entry)
    yield fd.find(argstr)

def _find_insensitive(entry, argstr):
    """Find regex in document (case insensitive): find.regex-i &lt;regex&gt;

Find text in the document that matches a given regular expression. The regular
expression syntax is that of python regular expressions. Matching dicards case."""
    fd = RegexFinder(entry, re.IGNORECASE)
    yield fd.find(argstr)

def replace(entry, findre, replstr=None):
    """Find/replace regex in document: find.replace &lt;find&gt; [&lt;replace&gt;]

Quickly find and replace phrases in the document using regular expressions.
The &lt;replace&gt; may contain backreferences to groups in the regular
expression using $0, $1, etc. In addition, a set of simple transformations
on these backreferences can be performed using instead the syntax ${n:t1[,t2,...]} where
&lt;n&gt is the backreference number and &lt;t1&gt;, &lt;t2&gt; are transformations:

u: upper case first character
U: upper case match
l: lower case first character
L: lower case match
t: title case match"""
    fd = RegexFinder(entry)
    yield fd.replace(findre, False, replstr)

def replace_i(entry, findre, replstr=None):
    """Find/replace regex in document (case insensitive): find.replace-i &lt;find&gt; [&lt;replace&gt;]

Quickly find and replace phrases in the document using regular expression.
The &lt;replace&gt; may contain backreferences to groups in the regular
expression using $0, $1, etc. In addition, a set of simple transformations
on these backreferences can be performed using instead the syntax ${n:t1[,t2,...]} where
&lt;n&gt is the backreference number and &lt;t1&gt;, &lt;t2&gt; are transformations:

u: upper case first character
U: upper case match
l: lower case first character
L: lower case match
t: title case matchs"""
    fd = RegexFinder(entry, re.IGNORECASE)
    yield fd.replace(findre, False, replstr)

def replace_all(entry, findre, replstr=None):
    """Find/replace all regex in document: find.regex.replace-all &lt;find&gt; [&lt;replace&gt;]

Quickly find and replace all phrases in the document using regular expressions.
The &lt;replace&gt; may contain backreferences to groups in the regular
expression using $0, $1, etc. In addition, a set of simple transformations
on these backreferences can be performed using instead the syntax ${n:t1[,t2,...]} where
&lt;n&gt is the backreference number and &lt;t1&gt;, &lt;t2&gt; are transformations:

u: upper case first character
U: upper case match
l: lower case first character
L: lower case match
t: title case match"""
    fd = RegexFinder(entry, 0)
    yield fd.replace(findre, True, replstr)

def replace_all_i(entry, findre, replstr=None):
    """Find/replace all regex in document: find.regex.replace-all-i &lt;find&gt; [&lt;replace&gt;]

Quickly find and replace all phrases in the document using regular expressions.
The &lt;replace&gt; may contain backreferences to groups in the regular
expression using $0, $1, etc. In addition, a set of simple transformations
on these backreferences can be performed using instead the syntax ${n:t1[,t2,...]} where
&lt;n&gt is the backreference number and &lt;t1&gt;, &lt;t2&gt; are transformations:

u: upper case first character
U: upper case match
l: lower case first character
L: lower case match
t: title case match"""
    fd = RegexFinder(entry, re.IGNORECASE)
    yield fd.replace(findre, True, replstr)

def replace_semantic(entry, findstr, replstr=None):
    """Find/replace in document (semantically): find.regex.replace-s &lt;find&gt; [&lt;replace&gt;]

Find and replace semantically"""
    fd = SemanticFinder(entry)
    yield fd.replace(findstr, False, replstr)

def replace_all_semantic(entry, findstr, replstr=None):
    """Find/replace in document (semantically): find.regex.replace-s &lt;find&gt; [&lt;replace&gt;]

Find and replace semantically"""
    fd = SemanticFinder(entry)
    yield fd.replace(findstr, True, replstr)

# ex:ts=4:et



# -*- coding: utf-8 -*-
#
#  edit.py - edit commander module
#
#  Copyright (C) 2010 - Jesse van den Kieboom
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import os
import glob
import sys
import types
import inspect

from gi.repository import GLib, Gio, Gedit

import commander.commands as commands
import commander.commands.completion
import commander.commands.result
import commander.commands.exceptions

__commander_module__ = True

@commands.autocomplete(filename=commander.commands.completion.filename)
def __default__(filename, view):
    """Edit file: edit &lt;filename&gt;"""

    doc = view.get_buffer()
    cwd = os.getcwd()

    if not doc.is_untitled():
        cwd = doc.get_file().get_location().get_parent().get_path()
    else:
        cwd = os.path.expanduser('~/')

    if not os.path.isabs(filename):
        filename = os.path.join(cwd, filename)

    matches = glob.glob(filename)
    files = []

    if matches:
        for match in matches:
            files.append(Gio.file_new_for_path(match))
    else:
        files.append(Gio.file_new_for_path(filename))

    if files:
        window = view.get_toplevel()
        Gedit.commands_load_locations(window, files, None, 0, 0)

    return commander.commands.result.HIDE

def _dummy_cb(num, total):
    pass

@commands.autocomplete(newfile=commander.commands.completion.filename)
def rename(view, newfile):
    """Rename current file: edit.rename &lt;newname&gt;"""

    doc = view.get_buffer()

    if not hasattr(doc, 'set_location'):
        raise commander.commands.exceptions.Execute('Your version of gedit does not support this action')

    if doc.is_untitled():
        raise commander.commands.exceptions.Execute('Document is unsaved and thus cannot be renamed')

    if doc.get_modified():
        raise commander.commands.exceptions.Execute('You have unsaved changes in your document')

    if not doc.get_file().is_local():
        raise commander.commands.exceptions.Execute('You can only rename local files')

    f = doc.get_file().get_location()

    if not f.query_exists(None):
        raise commander.commands.exceptions.Execute('Current document file does not exist')

    if os.path.isabs(newfile):
        dest = Gio.file_new_for_path(newfile)
    else:
        dest = f.get_parent().resolve_relative_path(newfile)

    if f.equal(dest):
        yield commander.commands.result.HIDE

    if not dest.get_parent().query_exists(None):
        # Check to create parent directory
        fstr, words, modifierret = (yield commands.result.Prompt('Directory does not exist, create? [Y/n] '))

        if fstr.strip().lower() in ['y', 'ye', 'yes', '']:
            # Create parent directories
            try:
                os.makedirs(dest.get_parent().get_path())
            except OSError as e:
                raise commander.commands.exceptions.Execute('Could not create directory')
        else:
            yield commander.commands.result.HIDE

    if dest.query_exists(None):
        fstr, words, modifierret = (yield commands.result.Prompt('Destination already exists, overwrite? [Y/n]'))

        if not fstr.strip().lower() in ['y', 'ye', 'yes', '']:
            yield commander.commands.result.HIDE

    try:
        f.move(dest, Gio.FileCopyFlags.OVERWRITE, None, _dummy_cb, None)

        doc.get_file().set_location(dest)
        yield commander.commands.result.HIDE
    except Exception as e:
        raise commander.commands.exceptions.Execute('Could not move file: %s' % (e,))

def _mod_has_func(mod, func):
    return func in mod.__dict__ and type(mod.__dict__[func]) == types.FunctionType

def _mod_has_alias(mod, alias):
    return '__root__' in mod.__dict__ and alias in mod.__dict__['__root__']

def _edit_command(view, mod, func=None):
    try:
        location = Gio.file_new_for_path(inspect.getsourcefile(mod))
    except:
        return False

    if not func:
        Gedit.commands_load_location(view.get_toplevel(), location, None, 0, 0)
    else:
        try:
            lines = inspect.getsourcelines(func)
            line = lines[-1]
        except:
            line = 0

        Gedit.commands_load_location(view.get_toplevel(), location, None, line, 0)

    return True

def _resume_command(view, mod, parts):
    if not parts:
        return _edit_command(view, mod)

    func = parts[0].replace('-', '_')

    if len(parts) == 1 and _mod_has_func(mod, func):
        return _edit_command(view, mod, mod.__dict__[func])
    elif len(parts) == 1 and _mod_has_alias(mod, parts[0]):
        return _edit_command(view, mod)

    if not func in mod.__dict__:
        return False

    if not commands.is_commander_module(mod.__dict__[func]):
        return False

    return _resume_command(view, mod.__dict__[func], parts[1:])

@commands.autocomplete(name=commander.commands.completion.command)
def command(view, name):
    """Edit commander command: edit.command &lt;command&gt;"""
    parts = name.split('.')

    for mod in sys.modules:
        if commands.is_commander_module(sys.modules[mod]) and (mod == parts[0] or _mod_has_alias(sys.modules[mod], parts[0])):
            if mod == parts[0]:
                ret = _resume_command(view, sys.modules[mod], parts[1:])
            else:
                ret = _resume_command(view, sys.modules[mod], parts)

            if not ret:
                raise commander.commands.exceptions.Execute('Could not find command: ' + name)
            else:
                return commander.commands.result.HIDE

    raise commander.commands.exceptions.Execute('Could not find command: ' + name)

COMMAND_TEMPLATE="""import commander.commands as commands
import commander.commands.completion
import commander.commands.result
import commander.commands.exceptions

__commander_module__ = True

def __default__(view, entry):
    \"\"\"Some kind of cool new feature: cool &lt;something&gt;

Use this to apply the cool new feature\"\"\"
    pass

# ex:ts=4:et
"""

def new_command(view, entry, name):
    """Create a new commander command module: edit.new-command &lt;command&gt;"""

    filename = os.path.join(GLib.get_user_config_dir(), 'gedit/commander/modules/' + name + '.py')

    if os.path.isfile(filename):
        raise commander.commands.exceptions.Execute('Commander module `' + name + '\' already exists')

    dirname = os.path.dirname(filename)

    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    f = open(filename, 'w')
    f.write(COMMAND_TEMPLATE)
    f.close()

    return __default__(filename, view)

def save(view):
    window = view.get_toplevel()
    Gedit.commands_save_document(window, view.get_buffer())

    return commander.commands.result.HIDE

def save_all(view):
    window = view.get_toplevel()
    Gedit.commands_save_all_documents(window)

    return commander.commands.result.HIDE

locals()['file'] = __default__
move = rename

# ex:ts=4:et



import gi
gi.require_version('Gedit', '3.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Peas', '1.0')
gi.require_version('PeasGtk', '1.0')

from .synctex import SynctexAppActivatable, SynctexWindowActivatable



# -*- coding: utf-8 -*-
#
#  Copyrignt (C) 2017 Jordi Mas <jmas@softcatala.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110-1301, USA.

import urllib.request, urllib.parse, urllib.error
import json
import locale
from .service import Service

class Apertium(Service):

    g_language_codes = []
    g_language_names = []
    g_locales_names = {}

    DEFAULT_LANGUAGE_NAMES = ["Spanish -> English",
            "English -> Spanish",
            "Catalan -> English",
            "English -> Catalan"]

    DEFAULT_LANGUAGE_CODES = ["spa|eng",
            "eng|spa",
            "cat|eng",
            "eng|cat",
    ]


    @staticmethod
    def _clean_for_ut():
        Apertium.g_language_codes = []
        Apertium.g_language_names = []
        Apertium.g_locales_names = {}

    def get_default_language_codes(self):
        return 'eng|spa'

    def has_api_key(self):
        return False

    def set_api_key(self, key):
        pass

    def set_server(self, server):
        self.server = server

    def get_api_hint(self):
        pass

    def init(self):
        self._fetch_remote_language_names_and_pairs()
   
    def get_language_names(self):
        if len(Apertium.g_language_codes) > 0 and len(Apertium.g_language_names) > 0:
            return Apertium.g_language_names

        return self.DEFAULT_LANGUAGE_NAMES

    def get_language_codes(self):
        if len(Apertium.g_language_codes) > 0 and len(Apertium.g_language_names) > 0:
            return Apertium.g_language_codes

        return self.DEFAULT_LANGUAGE_CODES

    def _get_lang_from_langcountry(self, language):
        if language is not None:
            index = language.find("_")
            if index > 0:
                language = language[:index]

        return language

    def _get_user_locale(self):
        user_locale = locale.getdefaultlocale()[0]
        user_locale = self._get_lang_from_langcountry(user_locale)
        if user_locale is None:
            user_locale = 'en'
        return user_locale

    def _get_language_name(self, langcode, locales_names):
        if langcode in locales_names:
            language_name = locales_names[langcode]
        else:
            locale_no_country = self._get_lang_from_langcountry(langcode)
            country = langcode[len(locale_no_country) + 1:]

            language_name = '{0} ({1})'.format(locales_names[locale_no_country],
                                               country)
        return language_name

    def get_language_pair_name(self, source, target, locales_names=None):
        if locales_names is None:
            locales_names = Apertium.g_locales_names

        source = self._get_language_name(source, locales_names)
        target = self._get_language_name(target, locales_names)
        return '{0} -> {1}'.format(source, target)

    def _is_langcode_in_list(self, langcode, locales_names):
        if langcode not in locales_names:
            locale_no_country = self._get_lang_from_langcountry(langcode)
            if locale_no_country not in locales_names:
                return False
        return True

    def _add_missing_locale_names_in_english(self, locales, locales_names):
        locales_names_en = self._get_remote_language_names(locales, 'en')
        for l in locales_names_en:
            if l not in locales_names:
                locales_names[l] = locales_names_en[l]

        return locales_names


    def _fetch_remote_language_names_and_pairs(self):

        if len(Apertium.g_language_names) > 0:
            return

        try:
            language_names = []
            en_names_requested = False

            user_locale = self._get_user_locale()
            language_pair_source, language_pair_target, locales, language_codes = self._get_remote_language_pairs()
            locales_names = self._get_remote_language_names(locales, user_locale)

            for i in range(len(language_pair_source)):
                source = language_pair_source[i]
                target = language_pair_target[i]

                if en_names_requested is False and \
                   (self._is_langcode_in_list(source, locales_names) or\
                   self._is_langcode_in_list(target, locales_names)):
                    en_names_requested = True
                    locales_names = self._add_missing_locale_names_in_english(locales, locales_names)

                language_pair = self.get_language_pair_name(source, target, locales_names)
                language_names.append(language_pair)

            Apertium.g_locales_names = locales_names
            Apertium.g_language_names = language_names
            Apertium.g_language_codes = language_codes

        except Exception as e:
            print("_fetch_remote_language_names_and_pairs exception {0}".format(e))


    def _get_remote_language_pairs(self):
        url = "{0}/listPairs".format(self.server)

        response = urllib.request.urlopen(url)
        data = json.loads(response.read().decode("utf-8"))
        pairs = data['responseData']

        locales = set()
        language_pair_source = []
        language_pair_target = []
        language_codes = []

        for pair in pairs:
            source = pair['sourceLanguage']
            target = pair['targetLanguage']
            language_pair = '{0}|{1}'.format(source, target)
            language_pair_source.append(source)
            language_pair_target.append(target)
            locales.add(source)
            locales.add(target)
            language_codes.append(language_pair)

        return language_pair_source, language_pair_target, locales, language_codes
 
    def _get_remote_language_names(self, locales, user_locale):
        locales_string = ''
        for locale_code in locales:
            locales_string += locale_code + '+'

        url = "{0}/listLanguageNames?locale={1}&languages={2}".format(self.server,
                user_locale, locales_string)

        response = urllib.request.urlopen(url)
        return json.loads(response.read().decode("utf-8"))

    def translate_text(self, text, language_pair):
        url = "{0}/translate?langpair={1}&markUnknown=no".format(self.server, language_pair)
        url += "&q=" + urllib.parse.quote_plus(text.encode('utf-8'))

        response = urllib.request.urlopen(url)
        data = json.loads(response.read().decode("utf-8"))
        translated = data['responseData']['translatedText']
        return translated

